cdef double _F_integrand(double t, void *data) nogil
cdef double _F_integrand1(double t, void *data) nogil
cdef double _F_integrand2(double t, void *data) nogil
cdef double _F_integrand3(double t, void *data) nogil
cdef double _F_integrand4(double t, void *data) nogil
